""" Transition System

A transition system is the tuple: <states, actions, tr, init, ap, labels>

A state is an integer number or a tuple of states (inductive definition)
An action is a string (so actions is a set of string)

tr is a transition relation State x Action x State, and a transition t is a tuple in tr

An atomic proposition is a Symbol (from PySMT)
labels map states to a list of labels.

Example of use:

# Creates the states
s1 = (1,)
s2 = (2,)

# Creates a transition system
t = TS(set([s1, s2]), set(["a","b"]),
       set([(s1,"a",s1),(s2,"b",s1)]),
       set([s1]),
       set([Symbol("i"),Symbol("f")]),
       {s1 : [Symbol("i")],
        s2 : [Symbol("f")]
       }
       )
"""

from __future__ import annotations
from collections.abc import Mapping, Iterator, Iterable
from typing import Type, Tuple, Optional, NewType, Collection, Union, List
from copy import deepcopy

from mc308.logics.bool import BoolFormula


from pysmt.shortcuts import (
    Symbol,
    TRUE,FALSE,
    Not,And,Or,Implies,Iff,
    is_valid
)
from pysmt.fnode import FNode

from functools import reduce, partial
import itertools

# Type definitions
class StateTuple(tuple):
    """ Tuples of states are useful to represent products """
    def __new__(self, *args : Union[int, StateTuple]):
        return tuple.__new__(StateTuple, args)
State =  Union[int, StateTuple]
Action = NewType('Action', str)

Trans = Tuple[State,Action,State]


class TS():
    def __init__(self : TS,
                 states : Collection[State],
                 actions : Collection[Action],
                 tr : Collection[Trans],
                 init : Collection[State],
                 ap : Collection[BoolFormula],
                 labels : Mapping[State,Collection[BoolFormula]]) -> None:
        """
        Creates a finite state transition system
        """

        self.states = set(states)
        self.actions = set(actions)
        self.tr = {}
        for (s,a,p) in tr:
            if not s in self.tr:
                self.tr[s] = set()
            self.tr[s].add((a,p))

        self.init = set(init)
        self.ap = set(ap)
        assert reduce(lambda x,y : x and y.is_symbol(),
                      ap, True)

        self.labels = {k : set() for k in self.states}
        for k,v in labels.items():
            self.labels[k] = set(v)

    def __eq__(self : TS, other):
        """
        Warning: shallow equality.
        """
        if not isinstance(other, TS):
            return False

        if not (self.states == other.states and
                self.actions == other.actions and
                self.tr == other.tr and
                self.init == other.init and
                self.ap == other.ap):
            return False

        for s in self.states:
            if self.labels[s] != other.labels[s]:
                return False
        return True


    def __repr__(self : TS) :
        pl = lambda l : ",".join(l)

        tr_list = set()
        for s in self.states:
            for (a,b) in self.tr.get(s, []):
                tr_list.add((s,a,b))

        return """TS : {
  states : %s;
  actions : %s;
  tr : %s;
  init : %s;
  ap : %s;
  labels : %s;
}""" % ( pl([str(l) for l in self.states]),
         pl(self.actions),
         pl(["(%s, %s, %s)" % (str(a[0]), a[1], str(a[2])) for a in tr_list]),
         pl([str(l) for l in self.init]),
         pl([str(a) for a in self.ap]),
         pl([ ("(%s, (%s))" % (str(a), ",".join([str(l) for l in v]))) for a,v in self.labels.items()]) )



    def eval_state(self : TS, s : State, formula : BoolFormula) -> bool:
        """
        Evaluate formula on the labels of the state s
        """
        assignment = {}
        labels = self.labels.get(s, set())
        for a in self.ap:
            assignment[a] = TRUE() if a in labels else FALSE()

        # If this assertion fails it means your formula has variables that are
        # not in the set of atomic propositions.
        assert(formula.get_free_variables() <= self.ap)

        res = formula.substitute(assignment).simplify()
        assert res == TRUE() or res == FALSE()
        return res == TRUE()

    def post(self : TS, s : State) -> List[Trans]:
        for (a,dst) in self.tr.get(s, []):
            yield (a,dst)

    def simulate(self : TS, run : FinitePath) -> bool:
        """
        Simulates an infinite run of TS.
        Returns true if that's the case, false otherwise.
        """
        trace = []
        if len(run) <= 0: return False

        if not run[0] in self.init:
            return False

        # check trans
        i = 0
        has_loop = False
        while (i + 1 < len(run)):
            found = False
            for (a, succ) in self.tr.get(run[i],[]):
                if succ == run[i+1]:
                    found = True
                    break
            if not found:
                return False
            i += 1

        return True


class FinitePath():
    """
    Represents a finite sequence of states
    """
    def __init__(self, path):
        self.path = [s for s in path]

    def add_state(self, state : State):
        self.path.append(state)

    def __len__(self):
        return len(self.path)

    def __getitem__(self, i):
        return self.path[i]

    def states(self):
        for a in self.path:
            yield a

    def __repr__(self):
        return " ".join([str(s) for s in self.path])


    def __eq__(self, other):
        return self.path == other.path


