""" Module for printing the transition system.

Uses the graphviz library to create a DOT file.
"""

import io
from typing import Collection, Tuple, BinaryIO

from collections.abc import Mapping

import graphviz

from mc308.auto.ts import TS




def get_dot(nodes : Collection[int],
            init_nodes : Collection[int],
            nodes_label : Mapping[int, str],
            edges : Collection[Tuple[int,int]],
            edges_labels : Mapping[Tuple[int,int], str]) -> graphviz.Graph:

    """ Create a dot representation for a graph """

    dot = graphviz.Digraph(comment='')

        # Add nodes with labels
    for node in nodes:
        label = nodes_label.get(node, str(node))  # Default to node ID if no label
        shape = "circle"
        dot.node(str(node), label=label, shape=shape)

        # Add initial states
        if node in sorted(list(init_nodes)):
            invisible_node = '_fake_init_%s' % str(node)
            dot.node(invisible_node, style="invisible")
            dot.edge('%s' % invisible_node, '%s' % str(node))

    # Add edges with labels
    for edge in edges:
        src, dst = edge
        label = edges_labels.get(edge, "")
        dot.edge(str(src), str(dst), label=label)

    return dot


def ts_to_dot(ts : TS) -> graphviz.Graph:
    """ Convert a transition system to a graph in the dot format """

    states_list = list(ts.states)
    m = {s: i for i, s in enumerate(states_list)}
    nodes = [m[s] for s in states_list]
    init_nodes = [m[s] for s in list(ts.init)]
    nodes_label = {m[s] : "%s\n{%s}" %(str(s), ",".join([str(a) for a in label])) for s,label in ts.labels.items()}

    edges= []
    edges_labels = {}
    for s in states_list:
        for (l, dst) in ts.tr.get(s, []):
            edge = (m[s], m[dst])
            edges.append(edge)
            edges_labels[edge] = l

    dot = get_dot(nodes, init_nodes, nodes_label, edges, edges_labels)
    return dot


def to_png(dot : graphviz.Graph,
           outstream : BinaryIO) -> None:
    png_bytes = dot._repr_image_png()
    outstream.write(png_bytes)




# def show_ts(ts : ts.TS, outstream : typing.BinaryIO) : -> None:

