"""
Parser for plain Boolean expression.

An atomic proposition (ap) is a standard identifier (regexp [a-zA-Z_][a-zA-Z_0-9]*).

A Boolean expression is defined as follows:

bexp := ap
     | True | False
     | ! bexp
     | bexp & bexp | bexp '|' bexp 
     | bexp -> bexp | bexp <-> bexp
     | (bexp)

"""

import ply.lex as lex
import ply.yacc as yacc

from pysmt.shortcuts import (
    Symbol,
    TRUE, FALSE,
    And, Or, Not, Implies
)

# Lexer

reserved = {
   'True' : 'TRUE',
   'False' : 'FALSE',
}

# List of token names
tokens = (
    'IDENTIFIER',
    'NOT',
    'AND', 'OR',
    'IMPLIES', 'IFF',
    'LPAREN', 'RPAREN'
) + tuple(reserved.values())


#t_TRUE = r'True'
#t_FALSE = r'False'
t_AND = r'\&'
t_OR = r'\|'
t_NOT = r'\!'
t_IMPLIES = r'->'
t_IFF = r'<->'
t_LPAREN = r'\('
t_RPAREN = r'\)'

# Regular expressions for the tokens

def t_IDENTIFIER(t):
    r'[a-zA-Z_][a-zA-Z_0-9]*'
    t.type = reserved.get(t.value,'IDENTIFIER')    # Check for reserved words
    return t

# Ignore whitespace
t_ignore = ' \t\n'

# from lower to higher
precedence = (
    ('right', 'IMPLIES', 'IFF'),
    ('right', 'AND', 'OR'),
    ('right', 'NOT'),
)

# Error handling rule
def t_error(t):
    print(f"Illegal character '{t.value[0]}'")
    t.lexer.skip(1)

# Build the lexer
bool_lexer = lex.lex()

# Parser
# Parsing rules
def p_expression_true(p):
    'expression : TRUE'
    p[0] = TRUE()

def p_expression_false(p):
    'expression : FALSE'
    p[0] = FALSE()

def p_expression_identifier(p):
    'expression : IDENTIFIER'
    p[0] = Symbol(p[1])

def p_expression_not(p):
    'expression : NOT expression'
    p[0] = Not(p[2])

def p_expression_and(p):
    'expression : expression AND expression'
    p[0] = And(p[1],p[3])

def p_expression_or(p):
    'expression : expression OR expression'
    p[0] = Or(p[1],p[3])

def p_expression_implies(p):
    'expression : expression IMPLIES expression'
    p[0] = Implies(p[1],p[3])

def p_expression_group(p):
    'expression : LPAREN expression RPAREN'
    p[0] = p[2]

# Error rule for syntax errors
def p_error(p):
    print("Syntax error in input!")

# Build the parser
bool_parser = yacc.yacc()


class BoolParser:
    def __init__(self):
        self.lexer = bool_lexer.clone()
        self.parser = bool_parser

    def parse(self, text):
        return self.parser.parse(text, lexer=self.lexer)

