import ply.lex as lex
import ply.yacc as yacc

from pysmt.shortcuts import (
    Symbol,
    TRUE, FALSE,
    And, Or, Not, Implies
)

# Lexer

reserved = {
   'True' : 'TRUE',
   'False' : 'FALSE',
   'U' : 'UNTIL',
   'F' : 'FINALLY',
   'G' : 'GLOBALLY',
   'X' : 'NEXT',
}

# List of token names
tokens = (
    'IDENTIFIER',
    'AND', 'OR', 'NOT', 'IMPLIES',
    'LPAREN', 'RPAREN'
) + tuple(reserved.values())


#t_TRUE = r'True'
#t_FALSE = r'False'
t_AND = r'\&'
t_OR = r'\|'
t_NOT = r'\!'
t_IMPLIES = r'->'
t_LPAREN = r'\('
t_RPAREN = r'\)'

# Regular expressions for the tokens

def t_IDENTIFIER(t):
    r'[a-zA-Z_][a-zA-Z_0-9]*'
    t.type = reserved.get(t.value,'IDENTIFIER')    # Check for reserved words
    return t

# Ignore whitespace
t_ignore = ' \t'

# from lower to higher
precedence = (
    ('right', 'IMPLIES', 'OR'),
    ('right', 'AND', 'OR'),
    ('right', 'UNTIL'),
    ('right', 'FINALLY', "GLOBALLY"),
    ('right', 'NOT'),
)

# Error handling rule
def t_error(t):
    print(f"Illegal character '{t.value[0]}'")
    t.lexer.skip(1)

# Build the lexer
lexer = lex.lex()

# Parser

# Parsing rules
def p_expression_true(p):
    'expression : TRUE'
    p[0] = TRUE()

def p_expression_false(p):
    'expression : FALSE'
    p[0] = FALSE()

def p_expression_identifier(p):
    'expression : IDENTIFIER'
    # Assign the truth value based on the identifier
    p[0] = Symbol(p[1])

def p_expression_not(p):
    'expression : NOT expression'
    p[0] = Not(p[2])

def p_expression_and(p):
    'expression : expression AND expression'
    p[0] = And(p[1],p[3])

def p_expression_or(p):
    'expression : expression OR expression'
    p[0] = Or(p[1],p[3])

def p_expression_implies(p):
    'expression : expression IMPLIES expression'
    p[0] = Implies(p[1],p[3])

def p_expression_group(p):
    'expression : LPAREN expression RPAREN'
    p[0] = p[2]

# Error rule for syntax errors
def p_error(p):
    print("Syntax error in input!")

# Build the parser
parser = yacc.yacc()

# Function to test the parser
def test_parser(expression):
    lexer.input(expression)
    result = parser.parse(expression)
    return result

# Examples of boolean expressions
if __name__ == '__main__':


    tc = [
        ("True", TRUE()),
        ("False", FALSE()),
        ("abc", Symbol("abc")),
        ("ab & b", And(Symbol("ab"), Symbol("b"))),
        ("True & False", And(TRUE(), FALSE())),
        ("! True | False", Or(Not(TRUE()), FALSE())),
        ("! a -> b", Implies(Not(Symbol("a")),Symbol("b")) ),
        ("! (a -> b)", Not(Implies(Symbol("a"),Symbol("b")) )),
        ]

    for a,b in tc:
        result = test_parser(a)

        print(result == b)
        if (result != b):
            print(f"Expected {b}")
            print(f"Result: {result}")
            print(f"Is true token: {result.is_true()}")
            print("")
        





