"""

Parse a transition system

states : 1,2,(3,(1,6));
actions : tau, tau2;
tr : (1,tau,2), (2,tau2,1);
init : 1, 2
ap : green, red
labels : (1, (green, red), ((3,(1,6)), := (green))

"""
import ply.lex as lex
import ply.yacc as yacc

from mc308.auto.ts import TS

from pysmt.shortcuts import (
    Symbol,
    TRUE, FALSE,
    And, Or, Not, Implies, Iff
)


reserved = {
   'states' : 'STATES_KEY',
   'actions' : 'ACTIONS_KEY',
   'tr' : 'TR_KEY',
   'init' : 'INIT_KEY',
   'ap' : 'AP_KEY',
   'labels' : 'LABELS_KEY',
}


# Lexer
tokens = (
    'IDENTIFIER', 'NUMBER',
    'COMMA', 'SEMI', 'COLON',
    'LPAREN', 'RPAREN'
)  + tuple(reserved.values())

t_COMMA = r','
t_SEMI = r';'
t_COLON = r':'
t_NUMBER = r'\d+'

t_LPAREN = r'\('
t_RPAREN = r'\)'

def t_IDENTIFIER(t):
    r'[a-zA-Z_][a-zA-Z_0-9]*'
    t.type = reserved.get(t.value,'IDENTIFIER')    # Check for reserved words
    return t

# Ignore whitespace
t_ignore = ' \t\n'


# Lexer error handling
def t_error(t):
    print(f"Illegal character '{t.value[0]}'")
    t.lexer.skip(1)

ts_lexer = lex.lex()

# Parser definition
def p_ts(p):
    '''ts : states_section init_section actions_section trans_section ap_section labels_section
    '''
    p[0] = {
        'states' : p[1],
        'init' : p[2],
        'actions' : p[3],
        'tr' : p[4],
        'ap' : p[5],
        'labels' : p[6],
    }

def p_states_section(p):
    '''states_section : STATES_KEY COLON states_list SEMI'''
    p[0] = p[3]

def p_actions_section(p):
    '''actions_section : ACTIONS_KEY COLON id_list SEMI'''
    p[0] = p[3]

def p_init_section(p):
    '''init_section : INIT_KEY COLON states_list SEMI'''
    p[0] = p[3]


def p_ap_section(p):
    '''ap_section : AP_KEY COLON id_list SEMI'''
    p[0] = p[3]


def p_labels_section(p):
    '''labels_section : LABELS_KEY COLON label_map SEMI'''
    p[0] = p[3]

def p_trans_section(p):
    '''trans_section : TR_KEY COLON trans_list SEMI'''
    p[0] = p[3]


# Parsing states section
def p_states_list(p):
    '''states_list : state
                   | states_list COMMA state
    '''
    if len(p) == 2:
        p[0] = [p[1]]
    else:
        p[0] = p[1] + [p[3]]

def p_state(p):
    '''state : NUMBER
             | LPAREN state COMMA state RPAREN
    '''
    if len(p) == 2:
        p[0] = (int(p[1]))

    else:
        p[0] = (p[2], p[4])

def p_id_list(p):
    '''id_list : IDENTIFIER
             | id_list COMMA IDENTIFIER
    '''
    p[0] = [p[1]] if len(p) == 2 else p[1] + [p[3]]

def p_trans_list(p):
    '''trans_list : tr
                  | trans_list COMMA tr
    '''
    p[0] = [p[1]] if len(p) == 2 else p[1] + [p[3]]

def p_tr(p):
    '''tr : LPAREN state COMMA IDENTIFIER COMMA state RPAREN
    '''
    p[0] = (p[2], p[4], p[6])


def p_label_map(p):
    '''label_map : label_map_element
             | label_map COMMA label_map_element
    '''
    p[0] = [p[1]] if len(p) == 2 else p[1] + [p[3]]

def p_label_map_element(p):
    '''label_map_element : LPAREN state COMMA LPAREN label_list RPAREN RPAREN
         | LPAREN state COMMA LPAREN RPAREN RPAREN
    '''
    if len(p) == 7:
        p[0] = (p[2], [])
    else:
        assert(len(p) == 8)
        p[0] = (p[2], p[5])


def p_label_list(p):
    '''label_list : IDENTIFIER
         | label_list COMMA IDENTIFIER
    '''
    p[0] = [p[1]] if len(p) == 2 else p[1] + [p[3]]




# Error rules
def p_error(p):
    if p:
        print(f"Syntax error at '{p.value}'")
    else:
        print("Syntax error at EOF")

# Build the parser
ts_parser = yacc.yacc()

class TSParser:
    def __init__(self):
        self.lexer = ts_lexer.clone()
        self.parser = ts_parser

    def _parse(self, text):
        return self.parser.parse(text, lexer=self.lexer)

    def parse(self, text):
        parsed = self._parse(text)

        if parsed is None:
            raise Exception("Parsing error")

        def check_state(states, s):
            if not s in states:
                msg = "State %s not contained in %s" % (str(s), ",".join([str(s) for s in states]))
                raise Exception(msg)

        # typecheck...

        states = set(parsed['states'])
        init = set(parsed['init'])
        for i in init: check_state(states,i)

        actions = set(parsed['actions'])

        tr = set()
        for (s1,a,s2) in parsed['tr']:
            check_state(states,s1)
            check_state(states,s2)
            check_state(actions,a)
            tr.add((s1,a,s2))

        ap = set([Symbol(a) for a in parsed['ap']])

        labels = {}
        for (i,label_list) in parsed['labels']:
            check_state(states,i)

            label_set = set([Symbol(l) for l in label_list])
            undecl = label_set - ap
            if len(undecl) > 0:
                msg = "Undeclared atomic propositions %s" % (",".join([str(s) for s in undecl]))
                raise Exception(msg)
            labels[i] = label_set

        ts = TS(states, actions, tr, init, ap, labels)
        return ts
